import pack.dogspack.*;
import pack.interfaces.*;


class MyGenericList<T> {

    T data;
    MyGenericList<T> next;

    MyGenericList() {
	data = null;
	next = null;
    }

}


class MyReallyGenericUnboudedStack<T> implements MyGenericStack<T> {

    MyGenericList<T> stack;
 
    int numElems;

    MyReallyGenericUnboudedStack() {

	stack = new MyGenericList<T>();

    }

    public void push(T o) {

	MyGenericList<T> x = new MyGenericList<T>();
	x.data = o;
	x.next = stack.next;
	stack.next = x;
	numElems++;

    }

    public T pop() {

	if ( stack.next != null ) {

	    T retval = stack.next.data;
	    stack.next = stack.next.next;
	    numElems--;
	    return retval;

	}
	else {
	    System.out.println("Can't pop from empty stack, stupid!");
	    return null;
	}
    }

    public boolean isEmpty() {
	return (numElems == 0);
    }


}




class MyReallyGenericStack<T>  implements MyGenericStack<T> {

    T[] boundedStack;
    int numElems;

    MyReallyGenericStack(int maxElems) {

        // Work-around fuer Erzeugung generischer Arrays:
        // Da nicht sicher ist, ob T einen Konstruktor
        // T(int maxElems) besitzt, wird ein Array ueber Object
        // erzeugt.
	boundedStack = (T[])new Object[maxElems];

	numElems = 0;

    }

    public void push(T o) {

	if ( numElems < boundedStack.length ) {
	    boundedStack[numElems++] = o;
	}
	else {
	    System.out.println("I'm full, buster!");
	}

    }

    public T pop() {

	if ( numElems > 0 ) {
	    return boundedStack[--numElems];
	}
	else {
	    System.out.println("Can't pop from empty stack, stupid!");
	    return null;
	}

    }

    public boolean isEmpty() {
	return (numElems == 0);
    }


}


class MyNotSoGenericStack implements MyStack {

    Object boundedStack[];
    int numElems;

    Class admissibleType;

    MyNotSoGenericStack(int maxElems, Class c) {

	numElems = 0;
	boundedStack = new Object[maxElems];
	admissibleType = c;

    }

    public void push(Object o) {

        // Alternativ: if ( o instanceof String)

	if ( o.getClass() == admissibleType  ) {

	    if ( numElems < boundedStack.length ) {
		boundedStack[numElems++] = o;
	    }
	    else {
		System.out.println("I'm full, buster!");
	    }

	}
	else {
	    System.out.println("This object is of evil type: " 
			       + o.toString());
	}

    }

    public Object pop() {

	if ( numElems > 0 ) {
	    return boundedStack[--numElems];
	}
	else {
	    System.out.println("Can't pop from empty stack, stupid!");
	    return null;
	}

    }

    public boolean isEmpty() {
	return (numElems == 0);
    }

}



public class MyGenerics {

    public static void main(String[] args) {

	MyNotSoGenericStack s0 
            = new MyNotSoGenericStack(10,String.class);

	s0.push("abc");
	s0.push("cde");
	s0.push(new Integer(19));

	String x = (String)s0.pop();
	System.out.println("pop'ed string is " + x);


	MyReallyGenericStack<String> s1 
	    = new MyReallyGenericStack<String>(10);


	s1.push("asdadd");
	s1.push("mmmmmmmmmmmmmmmm");
	// Wuerde eine Fehlermeldung des Compilers provozieren:
        //s1.push(new Integer(19));

	x = (String)s1.pop();
	System.out.println("pop'ed string is " + x);


	MyReallyGenericUnboudedStack<String> s2 
                    = new MyReallyGenericUnboudedStack<String>();

	s2.push("item0");
	s2.push("item1");
	s2.push("item2");
	s2.push("item3");
	// results in a compile error:
        // s2.push(new Integer(19));

	while ( (x = (String)s2.pop()) != null )
	    System.out.println("pop'ed string is " + x);

	MyNotSoGenericStack sDogs = new MyNotSoGenericStack(10,Dogs.class);
	sDogs.push(new Dogs("brown"));

        // Diese beiden Eintraege werden nicht akzeptiert,
        // da in push() genau auf Instanz von Dogs abgefragt wird.
	sDogs.push(new NiceDogs("white"));
	sDogs.push(new BadDogs("grey"));


        // In der Dogs-Instanz  des generischen Stacks werden
        // auch die NiceDogs-, BadDogs-Instanzen akzeptiert,
        // da diese auch vom Obertyp Dogs sind.
	MyReallyGenericStack<Dogs> sGenDogs = new MyReallyGenericStack<Dogs>(10);
	sGenDogs.push(new Dogs("brown"));
	sGenDogs.push(new NiceDogs("white"));
	sGenDogs.push(new BadDogs("grey"));

        // Da sDogs.pop() den Objekte vom Typ Object zurueck
        // gibt und Object keine Methode stroke() kennt,
        // muss vor dem Methodenaufruf eine cast-Operation erfolgen:
	while ( !sDogs.isEmpty() ) ((Dogs)sDogs.pop()).stroke();

        // Bei der Dogs-Instanz  des generischen Stacks 
        // koennen die stroke()-Operationen ihne vorherigen cast
        // erfolgen, da der Rueckgabetyp von pop() Dogs ist:
	while ( !sGenDogs.isEmpty() ) sGenDogs.pop().stroke();

    }


}
